/*
   podstawowe operacje na semaforach
   przy pokazaniu wzajemnego wykluczania sie
*/

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <errno.h>

#define SEMPERM 0x600
#define TRUE   1
#define FALSE  0

/*
   struktura, za pomoca ktorej zapisywane beda wartosci do semafora
   trzeba zadaklarowac, bo w wielu implementqacja unix'a
   nie ma tej struktury, np. w Solaris
*/

typedef union semun
{
	int wartosc;		/* wartosc semafora, l-ba calkowita dodatnia,
				/* ustawiona za pomoca funkcji systemowych, nigdy nie wprost*/
	struct semid_ds *bufor;
	ushort *tablica;        /* tablica liczb krotkich */
};

union semun arg;         /* ta struktura */

/* -------------------------------------------------------------------------
   Funkcja pomocnicza, tworzy 3-y zestawy semaforow o roznych warunkach i kluczach
   W pliku Init_sem.c jest ta sama funkcja.
*/

int Init_Semafora (key_t klucz semafora)
{
	int status =0, id_semafora;

	if ((id_semafora = semget(klucz_semafora, 1, SEMPERM | IPC_CREAT | IPC EXCL)) == -1 )
	{
		if (errno == EEXIST)
			id_semafora = semget (klucz_semafora, 1, 0);
	}
	else /* jesli semafor jest utworzony */
	{
		semun arg;
		arg.val = 1;
		status = semctl (id_semafora, 0, SETVAL, arg);
	}

	if (id_semafora == -1 || status == -1)  /* niepowodzenie */
	{
		perror("Utworzenie semafora nie powiodlo sie");
		return (-1);  /* wyjscie z programu z kodem bledu */
	}

	/*
	   jesli utworzenie sie powiodlo - dojdzie az tutaj
	   i funkcja Init_Semafora zwroci identyfikator semafora
	*/
	return (id_semafora);
}


/* -------------------------------------------------------------------------
   Funkcja odpowiadajaca opuszczeniu semafora: p()
*/

int p (int id_sem)
{
   struct sembuf p_bufor;   /*struktura sembuf jest str. systemowa !*/
   /*
      Kazdemu semaforowi w zestawie odpowiada taka struktura.
      W niej zapamietane sa operacje, ktore beda wykonane na semaforach.
   */
   p_bufor.sem_num = 0;        /* numer semafora w zestawie, numeracja od 0 */
   p_bufor.sem_op  = -1;       /* operacja wykonywana na semaforze tu : zmniejszanie jego wartosci */
   p_bufor.sem_flg = SEM_UNDO; /* ew. flagi */

   /*
      Parametry semop:
      - identyfikator semafora,
      - wskaznik na adres bazowy tablicy operacji semaforowych, one zostana wykonane na zest. semaforow
      - l-ba elementow tej tablicy operacji semaforowych
      Jesli operacja semop zakonczy sie niepowodzeniem - wypisz komunikat.
   */
   if (semop (id_sem, &p_bufor, 1) == -1)
   {
      perror("\n p ( %d ) - blad !", id_sem);
      exit(1);
   }
   return 0;
}

/* -------------------------------------------------------------------------
   Funkcja odpowiadajaca podniesieniu semafora: v()
*/

int v(int id_sem)
{
   struct sembuf v_bufor;   /* struktura sembuf jest struktura systemowa */

   v_bufor.sem_num = 0;        /* numer semafora w zestawie, numeracja od 0 */
   v_bufor.sem_op  = 1;        /* operacja wykonywana na semaforze tu : zmniejszanie jego wartosci */
   v_bufor.sem_flg = SEM_UNDO; /* ew. flagi */

   /* Jezeli operacja zakonczyla sie bledem - komunikat */
   if (semop(id_sem, &v_bufor, 1) == -1)
   {
      perror ("\n v( %d ) - blad !", id_sem);
      exit(1);
   }
   return 0;
}


/* ---------------------------------------------------------------------
   Funkcja usuwajaca zestaw semaforow
   Mozna ja sparametryzowac inczej, poprzez id_semafora
*/

   int Usn_Semafor (key_t klucz semafora)
   {
	int id_semafora;

        if ((id_semafora = semget(klucz_semafora, 1, SEMPERM | IPC_CREAT)) == -1 )
	{
		if (errno == EEXIST)
			id_semafora = semget (klucz_semafora, 1, 0);
	}
	else /* jesli semafor jest utworzony */
	{
        	if (semctl (id_semafora, 0, IPC_RMID, 0) == -1)
			perror("\n Blad usuwania zestawu semaforow ! ");  		
		else
			printf("\n Zestaw semaforow usuniety.");
		
	}
      
   }


/* ---------------------------------------------------------------------
   Funkcja, ktora:
   - za pomoca funkcji Init_Semafora tworzy zestaw semaforow (1 semafor w zestawie);
   - opuszcza semafor (zajmuje zasob krytyczny), za pomoca f-kcji p();
   - cos tam robi - gdy jest w sekcji krytycznej - tutaj usypia sie: sleep();
   - podnosi semafor (oddaje zasob), za pomoca funkcji v();
*/

void Uchwyc_Semafor (key_t klucz_semafora)
{
   int id_semafora;
   pid_t pid;

   pid = getpid();  /* pobiera identyfikator pid tego procesu, czyli swoj wlasny */

   /* Jesli blad utworzenia semafora - zakoncz */
   if ((id_semafora = Init_Semafora(klucz_semafora)) < 0)
      exit(1);

   printf("\n Proces %d przed sekcja krytyczna \n", pid);

   /* Proba opuszczenia semafora - zajecia zasobu krytycznego */
   p(id_semafora);
   printf(" \n !!! Proces %d w sekcji krytycznej !!! \n ", pid);
   sleep(10);
   printf("\n Proces %d zaraz opusci sekcje krytyczna \n", pid);

   /* podniesienie semafora */
   v(id_semafora);
   printf("\n Proces %d po sekcji krytycznej \n", pid);

   exit(0);
}


main ()
{
   key_t klucz_semafora;
   /*
      klucz semafora albo podany wprost, albo utworzyny za pomoca funkcji
      ftok, przyklad jej wywolania jest w programie rozne_s.c
   */
   klucz_semafora = 0x222;

   int i;  /* licznik petli */

   /*
      Progam tworzy za pomoca funkcji fork() trzy procesy - kopie siebie samego.
      Te procesy maja rozne identyfikatory pid. Wykonuja sie one (te nowo utworzone)
      od linijki z fork(). Dla kazdego z nich wywolywana
      jest funkcja Uchwyc_Semafor(), ktora tworzy semafor - jeden dla wszystkich,
      chociaz wydawac by sie moglo, ze 3-y razy. Klucz semafora bedzie taki sam,
      za pierwszym razem semafor bedzie utworzony, a potem tylko bedzie pobieramy
      jego identyfikator.
   */
   for(i=0; i<3; i++)
   {
      /* Jezeli prawiodlowe utworzenie procesu */
      if (fork () == 0)
      {
	 Uchwyc_Semafor(klucz_semafora);
      }
   }
}
